%let progname = ebv008.sas;
* xref:
* input:  ebvsim.sas
* output: e008
* does: Prentice(1988) EE for rho
        Efficiency for both rho and the log odds ratio
        Efficiency relative to GEE2
        Models for lambda_i: BB, MN, MAD
        A range of pi and rho
        Cluster size setup is chosen by a macro variable.
********************************************************************;
libname X "";
%let n = (2 : 100)` ;
%let n = (5 : 25 )` ;
%let n = {5  25  }` ;
%let seed = 13579;
********************************************************************;
proc iml
     worksize=30000  symsize=256; * <- adjust if necessary;

%include "ebvsim.sas";
********************************************************************;
start  var2cor (v);
/*
returns corr matrix
*/

  s = 1/sqrt(vecdiag(v));
  return ( s # v # t(s) );

finish;
********************************************************************;
start get_ee_b(b0, b1, b12, pi, rho);
/*
 Compute b0, b1, b12 (which define the EE)
 Prentice(1988)
*/
  psi= pi/(1-pi);
  a0 = psi-rho;   * y1+y2==0;
  a1 = -1-rho;    * y1+y2==1;
  a2 = 1/psi-rho; * y1+y2==2;
  b0 = a0;
  b1 = a1 - a0;
  b12= a0 + a2 - 2*a1;

finish;
********************************************************************;
start get_ee(y, n, pi, rho);
/*
 y = col vec of values
 n, pi, rho = scalars
 returns the ee as a matrix, #rows = nrow(y)
*/
  g1 = (y - n#pi) / (1+rho#(n-1));
  run get_ee_b(b0, b1, b12, pi, rho);
  g2 = (0.5*b0) # n#(n-1) + b1 # (n-1)#y + (0.5*b12) # y#(y-1)  ;
  return (g1 || g2);

finish;
********************************************************************;
start var_z12 (p, r);
/*
Returns var( b0 + b1*(Y1+Y2) + b12*Y1*Y2 )
Where E[Y1] = E[Y2] = p, corr(Y1,Y2)=r
*/

  run get_ee_b(b0, b1, b12, p, r);
  q = 1 - p;
  t = p*p + r*p*q;      * = E[Y1 Y2];
  return(
    2*p*q*(1+r)*b1*b1 + t*(1-t)*b12*b12 + 4*t*q*b1*b12
  );

finish;
********************************************************************;
start  derivative (d, n, p, r);
/*
ALR
- E[d g / d parms]
*/

  m = (n*n - n) /2 ;
  pq = p  * (1-p);
  c  = pq * (1-r) * (1-r) + r;
  d = j(2,2,0);
  d[1,1] = n / (1 + (n - 1) * r);
  /* d[1,2] = 0; */
  d[2,1] = m * (p+p-1) * r / c;
  d[2,2] = m * pq * (1+r) / ((1-r) * c);

finish;

********************************************************************;
start  mvc  (dist, mean, var, corr, n, pi, rho);
/*
Explicit formulae
*/

  if (upcase(dist)="MAD") then
    run mvc_Madsen   (mean, var, corr, n, pi, rho);
  else if (upcase(dist)="BB") then
    run mvc_BetaBin  (mean, var, corr, n, pi, rho);
  else if (upcase(dist)="MN") then
    run mvc_MorelNag (mean, var, corr, n, pi, rho);
  else do;
    print "Unrecognized distribution" dist;
    mean=0; var=0; corr=0;
    return;
  end;

finish;
********************************************************************;
start mean_var (dist, mean, var, corr, n, pi, rho);
/*
brute force through the pdf
*/

  if (upcase(dist)="MAD") then
    run madsen_d (cdf, pdf, n, pi, rho);
  else if (upcase(dist)="BB") then
    run betabin_d (cdf, pdf, n, pi, rho);
  else if (upcase(dist)="MN") then
    run morelnag_d (cdf, pdf, n, pi, rho);
  else do;
    print "Unrecognized distribution" dist;
    mean=0; var=0; corr=0;
    return;
  end;

  g = get_ee((0:n)`, n, pi, rho);
  pdf  = pdf`;
  mean = pdf * g;       * row vec;
  g    = g - repeat(mean, nrow(g), 1);
  var  = (g` # pdf) * g;
  corr = var2cor(var)[1,2];
  mean = mean`;

finish;
********************************************************************;
start mean_var3 (mean, var, corr, n, pi, rho);
/*
E[g] and var(g) under the 3-point distributuion
*/

  g = get_ee(0 // n*pi // n, n, pi, rho);
  run threept_d (cdf, pdf, n, pi, rho);
  pdf  = pdf`;
  mean = pdf * g;       * row vec;
  g    = g - repeat(mean, nrow(g), 1);
  var  = (g` # pdf) * g;
  corr = var2cor(var)[1,2];
  mean = mean`;

finish;


********************************************************************;
start mvc_Madsen (mean, var, corr, n, p, r);
/* E[g], var(g), corr(g1,g2) under the Madsen model */

  mean = j(2,1,0);
  var  = j(2,2,0);
  pq = p*(1-p) ;
  m  = (n*n-n)/2;
  var[2,2] = m * (1 + (m/pq - 3*m -1)*r - m*r*r);
  phi = 1 + (n-1) * r;
  var[1,2] = m * n * r * (1 - (p+p)) / phi;
  var[2,1] = var[1,2];
  var[1,1] = n * pq / phi;

  corr = var2cor(var)[1,2];

finish;

********************************************************************;
start mvc_BetaBin  (mean, var, corr, n, pi, rho);
/* E[g], var(g), corr(g1,g2) under the BetaBinomial  */

  mean = j(2,1,0);
  var  = j(2,2,0);
  pq  = pi*(1-pi) ;
  phi = 1+(n-1)*rho;
  a   = rho*(1+(3*n-4)*rho) +
          pq*(1+(n-4)*rho-3*(4*n-5)*rho*rho-n*rho*rho*rho);
  m   = (n*n - n)/2;
  num = m*phi*a;
  den = pq*(1+rho*(3+rho+rho));
  var[2,2] = num / den;
  var[1,2] = (m+m) * (1 - pi - pi) * rho / (1+rho);
  var[2,1] = var[1,2];
  var[1,1] = n * pq / phi;

  corr = var2cor(var)[1,2];

finish;

********************************************************************;
start mvc_MorelNag (mean, var, corr, n, pi, rho);
/* E[g], var(g), corr(g1,g2) under the Morel-Nagaraj */

  mean = j(2,1,0);
  var  = j(2,2,0);
  pq  = pi*(1-pi) ;
  phi = 1+(n-1)*rho;
  m   = (n*n - n)/2;

  s = sqrt(rho);
  a = 1 + 2*(n-4)*rho - 8*(n-2)*rho*s - (9-8*n+2*n*n)*rho*rho;
  num = rho*(2+4*(n-2)*s+(6-5*n+n*n)*rho) + 2*a*pq;
  num = num * m;
  den = pq + pq;

  var[2,2] = num / den;

  var[1,2] = m * rho * (1 - (pi+pi)) * (2+(n-2)*s) / phi;

  var[2,1] = var[1,2];
  var[1,1] = n * pq / phi;

  corr = var2cor(var)[1,2];

finish;

********************************************************************;
start chk_lambda(lambda, m);
  if (lambda < -1/(m-1)) then do;
    print "lambda = " lambda  " < -1/(m-1)";
  end;
  else if (lambda > 1)        then do;
    print "lambda = " lambda   " > 1";
  end;
finish;
********************************************************************;
start code_check(dummy);

  distribs = {"BB"  "MN"  "MAD"} ` ;
  do dist = 1 to nrow(distribs);
    distrib = distribs[dist];
    errors = 0;

    print
    "        n       pi        rho    c12         v11        v22   lambda";
    do pi  = 0.1 to 0.5 by 0.2;
      do n   = 2 to 40 by 2;
        do rho = 0 to 0.8 by 0.1;
          run mean_var  (distrib, mean, var, corr, n, pi, rho);
          v = vecdiag(var)` ;
          m = (n*n-n)/2;
          if (n>2) then do;
            vz12   = var_z12(pi, rho);
            lambda = (var[2,2]/(m*vz12) - 1)/(m-1);
            run chk_lambda(lambda, m);
          end;
          else lambda = .;
          * print n pi rho corr v lambda;

          run mvc  (distrib, mean1, var1, corr1, n, pi, rho);
          v1 = vecdiag(var1)` ;

          e      = sum(abs(mean1-mean)) + sum(abs(var1-var));
          norm   = sum(abs(mean1)) + sum(abs(var1));
          error  = (e > 1e-12 * norm);
          errors = errors +  error;
          if   (error) then
            print mean var, mean1 var1 ;

        end;
      end;
    end;
    print distrib errors "errors";
  end;

finish;

********************************************************************;
start get_delta (p, r);  /* derivative of (p,r)->(p,logOR) */
  d  = i(2);
  pq = p*(1-p);
  c  = pq * (1-r) * (1-r) + r;
  d[2,1] = (p+p-1)*r/(pq*c);
  d[2,2] = (1+r)/((1-r)*c);
  return (d);
finish;
********************************************************************;
start get_re (v1, v2);
/*
Ratios of diagonals
Returns a row vector (eff(pi) || eff(rho))
*/

  e = vecdiag(v2) / vecdiag(v1);

  return (e`);

finish;
********************************************************************;
start efficiency (dist, nn, p, r);
/*
nn := vector
p, r := params
dist := distrib name
*/

  ef_opt = j(2,2,0);
  sum_d   = j(2,2,0);
  sum_v   = j(2,2,0);
  do i = 1 to nrow(nn);
    n = nn[i];
    run mvc  (dist, mean, v, c, n, p, r);
    run  derivative (d, n, p, r);
    ef_opt = ef_opt + d` * inv(v) * d;
    sum_d = sum_d + d;
    sum_v = sum_v + v;
  end;

  ef_p88 = sum_d` * inv(sum_v) * sum_d;
  av_opt_rho = inv(ef_opt);
  av_p88_rho = inv(ef_p88);

  /* delta method */
  delta  = get_delta (p, r); /* derivative of (p,r)->(p,logOR) */
  av_p88_lor = delta * av_p88_rho * delta`;
  av_opt_lor = delta * av_opt_rho * delta`;

  ae_p88_rho = get_re (av_p88_rho, av_opt_rho);
  ae_p88_lor = get_re (av_p88_lor, av_opt_lor);
  return  (ae_p88_rho || ae_p88_lor[2]); /* order: pi rho lor */

finish;

********************************************************************;
/* iml execution starts here -> */

*reset noname;
*run code_check(0);
*endsas;

datalbl  = {dist pi rho ae_p88_pi ae_p88_rho ae_p88_lor};
datasave = j(1, ncol(datalbl), 0);
create  X.E008  from datasave [colname=datalbl];
setout  X.E008;     * output data set;

n  = &n;
n  = shape(n, nrow(n)*ncol(n), 1); /* make sure it is a column */
print n ;

rho_vals = (0:100)/1000 || (11:99)/100;
n_rho_vals = nrow(rho_vals)*ncol(rho_vals);

distribs = {"BB"  "MN"  "MAD"} ` ;

do dist = 1 to nrow(distribs);
  distrib = distribs[dist];
  do pi  = 0.1 to 0.5 by 0.1;
      do rho_id = 1 to n_rho_vals;
        rho  = rho_vals[rho_id];
        rho  = round(rho, 1e-3);
        pi   = round(pi , 1e-2);
        ae_p88 = efficiency (distrib, n, pi, rho);
        datasave = dist || pi || rho || ae_p88;
        append from  datasave;
       *print distrib pi rho ae_p88;
     end;
  end;
end;

close   X.E008;     * output data set;
********************************************************************;
