%let progname = ebv010.sas;
* xref:
* input:  ebvsim.sas
* output: e010
* does: Pseudo-likelihood
        Efficiency for both rho and the log odds ratio
        Efficiency relative to GEE2
        Models for lambda_i: BB, MN, MAD
        A range of pi and rho
        Cluster size setup is chosen by a macro variable.
********************************************************************;
libname X "";
%let n = (2 : 100)` ;
%let n = (5 : 25 )` ;
%let n = {5  25  }` ;
%let seed = 13579;
********************************************************************;
proc iml
     worksize=30000  symsize=512; * <- adjust if necessary;

%include "ebvsim.sas";
********************************************************************;
start  var2cor (v);
/*
returns corr matrix
*/

  s = 1/sqrt(vecdiag(v));
  return ( s # v # t(s) );

finish;
********************************************************************;
start get_ee_a(b0, b1, b12, pi, rho);
/*
Pseudo-likelihood ee for pi
Compute b0, b1, b12 (which define the EE)
*/
  a  = 1   - pi * (1 - rho);
  b  = rho + pi * (1 - rho);
  a0 = (rho / a - 2) / (1 - pi);            * y1+y2==0;
  a1 = (1 - pi - pi) / (pi * (1 - pi));     * y1+y2==1;
  a2 = (2 - rho / b) / pi;                  * y1+y2==2;
  b0 = a0;
  b1 = a1 - a0;
  b12= a0 + a2 - a1 - a1;

finish;
********************************************************************;
start get_ee_b(b0, b1, b12, pi, rho);
/*
Pseudo-likelihood ee for rho
Compute b0, b1, b12 (which define the EE)
*/
  psi= pi/(1-pi);
  a0 = 1/(1/psi+rho);  * y1+y2==0;
  a1 = -1/(1-rho);     * y1+y2==1;
  a2 = 1/(psi+rho);    * y1+y2==2;
  b0 = a0;
  b1 = a1 - a0;
  b12= a0 + a2 - 2*a1;

finish;
********************************************************************;
start get_ee(y, n, pi, rho);
/*
 y = col vec of values
 n, pi, rho = scalars
 returns the ee as a matrix, #rows = nrow(y)
*/
  nn1 = n # (n-1);
  yn1 = y # (n-1);
  yy1 = y # (y-1);

  * ee for pi;
  run get_ee_a(b0, b1, b12, pi, rho);
  g1 = (0.5*b0) # nn1 + b1 # yn1 + (0.5*b12) # yy1  ;

  * ee for rho;
  run get_ee_b(b0, b1, b12, pi, rho);
  g2 = (0.5*b0) # nn1 + b1 # yn1 + (0.5*b12) # yy1  ;

  return (g1 || g2);

finish;
********************************************************************;
start var_z12 (p, r);
/*
Returns var( b0 + b1*(Y1+Y2) + b12*Y1*Y2 )
Where E[Y1] = E[Y2] = p, corr(Y1,Y2)=r
*/

  run get_ee_b(b0, b1, b12, p, r);
  q = 1 - p;
  t = p*p + r*p*q;      * = E[Y1 Y2];
  return(
    2*p*q*(1+r)*b1*b1 + t*(1-t)*b12*b12 + 4*t*q*b1*b12
  );

finish;
********************************************************************;
start  derivative (d, n, p, r);
/*
Pseudo-likelihood
- E[d g / d parms]
*/

  d  = i(2);
  pq = p * (1-p);
  a  = pq * (1-r) * (1-r) + r;

  x = r * (2-r) + pq * (2 - 6*r + 4*r*r);
  y = pq * a;
  d[1,1] = x / y;

  d[1,2] = (p + p - 1) * r / a;
  d[2,1] = d[1,2];

  d[2,2] = (pq / a) * ( (1+r) / (1-r) );
  * so far, this D is for one (j,k) pair;

  m = (n*n - n) / 2 ;
  d = m * d;    * for n_choose_2 pairs;

finish;

********************************************************************;
start  mvc  (dist, mean, var, corr, n, pi, rho);
/*
Explicit formulae
*/

  if (upcase(dist)="BB") then
    run mvc_BetaBin  (mean, var, corr, n, pi, rho);
  else if (upcase(dist)="MN") then
    run mvc_MorelNag (mean, var, corr, n, pi, rho);
  else if (upcase(dist)="MAD") then
    run mvc_Madsen   (mean, var, corr, n, pi, rho);
  else do;
    print "Unrecognized distribution" dist;
    mean=.; var=.; corr=.;
    return;
  end;

finish;
********************************************************************;
start mean_var (dist, mean, var, corr, n, pi, rho);
/*
brute force through the pdf
*/

  if (upcase(dist)="BB") then
    run betabin_d (cdf, pdf, n, pi, rho);
  else if (upcase(dist)="MN") then
    run morelnag_d (cdf, pdf, n, pi, rho);
  else if (upcase(dist)="MAD") then
    run madsen_d (cdf, pdf, n, pi, rho);
  else do;
    print "Unrecognized distribution" dist;
    mean=.; var=.; corr=.;
    return;
  end;

  g = get_ee((0:n)`, n, pi, rho);
  pdf  = pdf`;
  mean = pdf * g;       * row vec;
  g    = g - repeat(mean, nrow(g), 1);
  var  = (g` # pdf) * g;
  corr = var2cor(var)[1,2];
  mean = mean`;

finish;
********************************************************************;
start mean_var3 (mean, var, corr, n, pi, rho);
/*
E[g] and var(g) under the 3-point distributuion
*/

  g = get_ee(0 // n*pi // n, n, pi, rho);
  run threept_d (cdf, pdf, n, pi, rho);
  pdf  = pdf`;
  mean = pdf * g;       * row vec;
  g    = g - repeat(mean, nrow(g), 1);
  var  = (g` # pdf) * g;
  corr = var2cor(var)[1,2];
  mean = mean`;

finish;


********************************************************************;
start mvc_BetaBin  (mean, v, corr, n, p, r);
/* E[g], var(g), corr(g1,g2) under the BetaBinomial  */
/* Pseudo-likelihood */

  mean = j(2,1,0);
  v    = j(2,2,0);
  m    = (n*n-n)/2;
  pq   = p*(1-p);
  r2   = r * r;
  a = r * ((2+r+r-r2)*n - 2 - r);
  a = a + (pq+pq) * (1-r) * ((1-r2-r2) * n - 1);
  d = pq * (1-r) * (1-r) + r;
  b = d * (1+r) * (1+r+r);
  c = 1 + (n+n-1)*r + n*(n-1)*r2;

  v[1,1] = m * a * (1+(n-1)*r) / (pq * b);

  v[1,2] = m * r * (p+p-1) * c / b;

  v[2,2] = m * pq * c / ((1-r)*(1+r+r)*d);

  v[2,1] = v[1,2];
  corr   = var2cor(v)[1,2];

finish;

********************************************************************;
start mvc_BetaBin1 (mean, v, corr, n, p, r);
/* original "raw" code  */
/* E[g], var(g), corr(g1,g2) under the BetaBinomial  */
/* Pseudo-likelihood */

  mean = j(2,1,0);
  v    = j(2,2,0);

v[1,1]=
((-1 + n)*n*(1 + (-1 + n)*r)*(r*(2 + r + n*(-2 - 2*r + r**2)) -
      2*p*(-1 + r)*(1 + n*(-1 + 2*r**2)) +
      2*p**2*(-1 + r)*(1 + n*(-1 + 2*r**2))))/
  (2*(-1 + p)*p*(p*(-1 + r)**2 - p**2*(-1 + r)**2 + r)*(1 + 3*r + 2*r**2))
;

v[1,2]=
((-1 + n)*n*(-1 + 2*p)*r*(1 + (-1 + 2*n)*r + (-1 + n)*n*r**2))/
  (2*(p*(-1 + r)**2 - p**2*(-1 + r)**2 + r)*(1 + 3*r + 2*r**2))
;


v[2,2]=
((-1 + n)*n*(-1 + p)*p*(1 + (-1 + 2*n)*r + (-1 + n)*n*r**2))/
  (2*(p*(-1 + r)**2 - p**2*(-1 + r)**2 + r)*(-1 - r + 2*r**2))
;

  v[2,1] = v[1,2];
  corr = var2cor(v)[1,2];

finish;

********************************************************************;
start mvc_MorelNag (mean, v, corr, n, p, r);
/* E[g], var(g), corr(g1,g2) under the Morel-Nagaraj */
/* Pseudo-likelihood */

  mean = j(2,1,0);
  v    = j(2,2,0);
  s = sqrt(r);

v[1,1]=
-(n*(r**2*(4 - 14*r + n**3*(-2 + s)**2*r + 16*r**1.5 - 6*r**2 +
          n**2*(4 - 20*r + 20*r**1.5 - 6*r**2) +
          n*(-8 + 30*r - 32*r**1.5 + 11*r**2)) -
       8*p**3*(-1 + s)**3*
        (-1 - 3*s + r + 11*r**1.5 + 2*r**2 - 18*r**2.5 +
          n**2*(-1 - 3*s + 3*r + 17*r**1.5 + 4*r**2 - 20*r**2.5) +
          n**3*(-r - 3*r**1.5 + 4*r**2.5) +
          n*(2 + 6*s - 3*r - 25*r**1.5 - 6*r**2 + 34*r**2.5)) +
       4*p**4*(-1 + s)**3*
        (-1 - 3*s + r + 11*r**1.5 + 2*r**2 - 18*r**2.5 +
          n**2*(-1 - 3*s + 3*r + 17*r**1.5 + 4*r**2 - 20*r**2.5) +
          n**3*(-r - 3*r**1.5 + 4*r**2.5) +
          n*(2 + 6*s - 3*r - 25*r**1.5 - 6*r**2 + 34*r**2.5)) -
       2*p*r*(-4 + 21*r - 4*r**1.5 - 52*r**2 + 60*r**2.5 - 21*r**3 +
          2*n**3*r*(-2 + s + 6*r - 7*r**1.5 + 2*r**2) -
          2*n**2*(2 - 14*r + 4*r**1.5 + 33*r**2 - 36*r**2.5 + 11*r**3) +
          n*(8 - 45*r + 10*r**1.5 + 106*r**2 - 118*r**2.5 + 39*r**3)) +
       2*p**2*(2 - 18*r + 71*r**2 - 20*r**2.5 - 150*r**3 + 172*r**3.5 -
          57*r**4 + 2*n**3*(r - 8*r**2 + 5*r**2.5 + 15*r**3 - 19*r**3.5 +
             6*r**4) - 2*n**2*(-1 + 11*r - 49*r**2 + 20*r**2.5 + 88*r**3 -
             100*r**3.5 + 31*r**4) +
          n*(-4 + 38*r - 153*r**2 + 50*r**2.5 + 296*r**3 - 334*r**3.5 +
             107*r**4))))/
  (4*(-1 + p)*p*(1 + p*(-1 + r))**2*(p + r - p*r)**2)
;


v[1,2]=
-((-1 + n)*n*(-1 + 2*p)*r*(r*(-2 + 2*(5 - 5*n + n**2)*s +
          (-10 + 11*n - 3*n**2)*r + (6 - 5*n + n**2)*r**1.5) -
       2*p*(-1 + s)**2*(1 + (1 + 3*n - n**2)*s -
          (3 - 5*n + n**2)*r + (9 - 8*n + 2*n**2)*r**1.5) +
       2*p**2*(-1 + s)**2*
        (1 + (1 + 3*n - n**2)*s - (3 - 5*n + n**2)*r +
          (9 - 8*n + 2*n**2)*r**1.5)))/
  (4*(1 + p*(-1 + r))**2*(1 + s)*(p + r - p*r)**2)
;


v[2,2]=
-((-1 + n)*n*(-1 + p)*p*(r*(-2 + (6 - 4*n)*s - (8 - 5*n + n**2)*r +
          (16 - 15*n + 3*n**2)*r**1.5 + (-10 + 11*n - 3*n**2)*r**2 +
          (6 - 5*n + n**2)*r**2.5) -
       2*p*(-1 + s)**2*(1 + 3*s + 2*n*r - 2*(-4 + n)*r**1.5 +
          (-5 + 8*n - 2*n**2)*r**2 + (9 - 8*n + 2*n**2)*r**2.5) +
       2*p**2*(-1 + s)**2*
        (1 + 3*s + 2*n*r - 2*(-4 + n)*r**1.5 +
          (-5 + 8*n - 2*n**2)*r**2 + (9 - 8*n + 2*n**2)*r**2.5)))/
  (4*(1 + p*(-1 + r))**2*(-1 + s)*(1 + s)**2*(p + r - p*r)**2)
;


  v[2,1] = v[1,2];
  corr = var2cor(v)[1,2];

finish;

********************************************************************;
start mvc_Madsen (mean, v, corr, n, p, r);
/* E[g], var(g), corr(g1,g2) under the Madsen model */
/* Pseudo-likelihood */

  mean = j(2,1,0);
  v    = j(2,2,0);

v[1,1]=
-((-1 + n)*n*((-1 + n)*r**2*(4 + (-4 + n)*r) -
       p*(-1 + r)*r*(-8 + 26*r + n**2*r*(4 + r) - n*(-8 + 28*r + r**2)) -
       8*p**3*(-1 + r)*(1 - 6*r + 11*r**2 + n**2*r*(-1 + r + r**2) -
          n*(1 - 7*r + 10*r**2 + r**3)) +
       4*p**4*(-1 + r)*(1 - 6*r + 11*r**2 + n**2*r*(-1 + r + r**2) -
          n*(1 - 7*r + 10*r**2 + r**3)) +
       p**2*(-1 + r)*(4 - 32*r + 70*r**2 + n**2*r*(-4 + 8*r + 5*r**2) -
          n*(4 - 36*r + 68*r**2 + 5*r**3))))/
  (4*(-1 + p)*p*(1 + p*(-1 + r))**2*(p + r - p*r)**2)
;


v[1,2]=
-((-1 + n)*n*(-1 + 2*p)*r*((4 - 5*n + n**2)*r +
       p**2*(-2 + 14*r + n**2*(-2 + 3*r + r**2) - n*(-6 + 15*r + r**2)) +
       p*(2 - 14*r - n**2*(-2 + 3*r + r**2) + n*(-6 + 15*r + r**2))))/
  (4*(p*(-1 + r)**2 - p**2*(-1 + r)**2 + r)**2)
;


v[2,2]=
-((-1 + n)*n*(-1 + p)*p*((-1 + n)*(n*(-1 + r) - 4*r)*r -
       p*(2 + (4 + 3*n - 3*n**2)*r + 2*(9 - 11*n + 3*n**2)*r**2 +
          (-1 + n)*n*r**3) + p**2*
        (2 + (4 + 3*n - 3*n**2)*r + 2*(9 - 11*n + 3*n**2)*r**2 +
          (-1 + n)*n*r**3)))/
  (4*(1 + p*(-1 + r))**2*(-1 + r)*(p + r - p*r)**2)
;


  v[2,1] = v[1,2];
  corr = var2cor(v)[1,2];

finish;

********************************************************************;
start chk_lambda(lambda, m);
  if (lambda < -1/(m-1)) then do;
    print "lambda = " lambda  " < -1/(m-1)";
  end;
  else if (lambda > 1)        then do;
    print "lambda = " lambda   " > 1";
  end;
finish;
********************************************************************;
start code_check(dummy);

  distribs = {"BB"  "MN"  "MAD"} ` ;
  do dist = 1 to nrow(distribs);
    distrib = distribs[dist];
    errors = 0;

    print
    "        n       pi        rho    c12         v11        v22   lambda";
    do pi  = 0.1 to 0.5 by 0.2;
      do n   = 2 to 40 by 2;
        do rho = 0 to 0.8 by 0.1;
          run mean_var  (distrib, mean, var, corr, n, pi, rho);
          v = vecdiag(var)` ;
          m = (n*n-n)/2;
          if (n>2) then do;
            vz12   = var_z12(pi, rho);
            lambda = (var[2,2]/(m*vz12) - 1)/(m-1);
            run chk_lambda(lambda, m);
          end;
          else lambda = .;
          * print n pi rho corr v lambda;

          run mvc  (distrib, mean1, var1, corr1, n, pi, rho);
          v1 = vecdiag(var1)` ;

          e      = sum(abs(mean1-mean)) + sum(abs(var1-var));
          norm   = sum(abs(mean1)) + sum(abs(var1));
          error  = (e > 1e-12 * norm);
          errors = errors +  error;
          if   (error) then
            print mean var, mean1 var1 ;

        end;
      end;
    end;
    print distrib errors "errors";
  end;

finish;

********************************************************************;
start get_delta (p, r);  /* derivative of (p,r)->(p,logOR) */
  d  = i(2);
  pq = p*(1-p);
  c  = pq * (1-r) * (1-r) + r;
  d[2,1] = (p+p-1)*r/(pq*c);
  d[2,2] = (1+r)/((1-r)*c);
  return (d);
finish;
********************************************************************;
start get_re (v1, v2);
/*
Ratios of diagonals
Returns a row vector (eff(pi) || eff(rho))
*/

  e = vecdiag(v2) / vecdiag(v1);

  return (e`);

finish;
********************************************************************;
start efficiency (dist, nn, p, r);
/*
nn := vector
p, r := params
dist := distrib name
*/

  ef_opt = j(2,2,0);
  sum_d   = j(2,2,0);
  sum_v   = j(2,2,0);
  sum_hd_ch  = j(2,2,0);
  sum_hvh_ch = j(2,2,0);
  h_kn = {1, 1};
  sum_hd_kn  = j(2,2,0);
  sum_hvh_kn = j(2,2,0);
  do i = 1 to nrow(nn);
    n = nn[i];
    run mvc  (dist, mean, v, c, n, p, r);
    run  derivative (d, n, p, r);
    ef_opt = ef_opt + d` * inv(v) * d;
    sum_d = sum_d + d;
    sum_v = sum_v + v;

    /* h_ch    = i(2)/(n-1); * le C & van H (1994);  */
    sum_hd_ch  = sum_hd_ch  + d / (n-1);
    sum_hvh_ch = sum_hvh_ch + v / ((n-1)*(n-1));

    h_kn[1]    = 1/(n-1);   * K & N (2000) weigh the PL U_beta only;
    sum_hd_kn  = sum_hd_kn  + h_kn # d;
    sum_hvh_kn = sum_hvh_kn + h_kn # v # h_kn`;
  end;

  av_opt_rho = inv(ef_opt);

  ef_pl      = sum_d` * inv(sum_v) * sum_d;
  av_pl_rho  = inv(ef_pl);

  ef_ch      = sum_hd_ch` * inv(sum_hvh_ch) * sum_hd_ch;
  av_ch_rho  = inv(ef_ch);

  ef_kn      = sum_hd_kn` * inv(sum_hvh_kn) * sum_hd_kn;
  av_kn_rho  = inv(ef_kn);


  /* delta method */
  delta      = get_delta (p, r); /* derivative of (p,r)->(p,logOR) */

  av_opt_lor = delta * av_opt_rho * delta`;
  av_pl_lor  = delta * av_pl_rho  * delta`;
  av_ch_lor  = delta * av_ch_rho  * delta`;
  av_kn_lor  = delta * av_kn_rho  * delta`;

  ae_pl_rho  = get_re (av_pl_rho, av_opt_rho);
  ae_pl_lor  = get_re (av_pl_lor, av_opt_lor);
  ae_ch_rho  = get_re (av_ch_rho, av_opt_rho);
  ae_ch_lor  = get_re (av_ch_lor, av_opt_lor);
  ae_kn_rho  = get_re (av_kn_rho, av_opt_rho);
  ae_kn_lor  = get_re (av_kn_lor, av_opt_lor);

  return  (   /* order: pi rho lor; PL  C&H  K&N */
    ae_pl_rho || ae_pl_lor[2] ||
    ae_ch_rho || ae_ch_lor[2] ||
    ae_kn_rho || ae_kn_lor[2]
  );

finish;

********************************************************************;
/* iml execution starts here -> */

*reset noname;
*run code_check(0);
*endsas;

datalbl  = {
  dist pi rho
  ae_pl_pi ae_pl_rho ae_pl_lor
  ae_ch_pi ae_ch_rho ae_ch_lor
  ae_kn_pi ae_kn_rho ae_kn_lor
};
datasave = j(1, ncol(datalbl), 0);
create  X.E010  from datasave [colname=datalbl];
setout  X.E010;     * output data set;

n  = &n;
n  = shape(n, nrow(n)*ncol(n), 1); /* make sure it is a column */
print n ;

rho_vals = (0:100)/1000 || (11:99)/100;
n_rho_vals = nrow(rho_vals)*ncol(rho_vals);

distribs = {"BB"  "MN"  "MAD"} ` ;

do dist = 1 to nrow(distribs);
  distrib = distribs[dist];
  do pi  = 0.1 to 0.5 by 0.1;
      do rho_id = 1 to n_rho_vals;
        rho  = rho_vals[rho_id];
        rho  = round(rho, 1e-3);
        pi   = round(pi , 1e-2);
        ae   = efficiency (distrib, n, pi, rho);
        datasave = dist || pi || rho || ae;
        append from  datasave;
       *print distrib pi rho ae;
      end;
  end;
end;

close   X.E010;     * output data set;
********************************************************************;
